﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Reflection;
using UnityEngine;

using ColossalFramework.Plugins;//temp

class Util
{   

    static public bool containsOneOf(string str, string[] substrs)
    {
        if (substrs.Length == 0)
            return true;
        for (int i = 0; i < substrs.Length; ++i)
            if (str.Contains(substrs[i].Trim()))
                return true;
        return false;
    }


    static public System.Type[] findConcreteSubclasses(System.Type baseType)
    {
        //get all concrete subclasses of baseType among all loaded assemblies
        var toolTypes = new List<System.Type>();
        foreach (Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
            try
            {
                foreach (Type type in assembly.GetTypes().Where(myType => myType.IsClass && !myType.IsAbstract && myType.IsSubclassOf(baseType)))
                    toolTypes.Add(type);
            }
            catch(Exception e)
            {
                Debug.Log(e.Message);
            }

        //DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, type);
               
     
        return toolTypes.ToArray();
    }

    public const BindingFlags ALL_METHOD_FLAGS = BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.Public;
    
    public static void tryMethodRedirection(System.Type type, string methodName,
                                            System.Type targetType, string targetMethodName,
                                            ref Dictionary<MethodInfo, Detours.RedirectCallsState> redirects,
                                            int paramCount = -1)
    {
        
        try
        {
            MethodInfo method = paramCount > -1 ? type.GetMethods(ALL_METHOD_FLAGS).Single(c => c.Name == methodName && c.GetParameters().Length == paramCount)
                                                : type.GetMethod(methodName, ALL_METHOD_FLAGS);
            if (method != null)
            {
                redirects.Add(method, Detours.RedirectionHelper.RedirectCalls(method, targetType.GetMethod(targetMethodName, ALL_METHOD_FLAGS)));
                //Debug.Log("Detoured " + type.Name+"::"+methodName + (paramCount > -1 ? " (" + paramCount + " params)" : ""));
            }
        }
        catch (Exception e)
        {
            Debug.Log("Exception detouring " + type.Name+"::"+methodName + (paramCount > -1 ? " (" + paramCount + " params)" : "") + ": " + e.Message);
            DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "Exception detouring " + type.Name + "::" + methodName + (paramCount > -1 ? " (" + paramCount + " params)" : "") + ": " + e.Message);
        };
    }

    //to quickly create generic semi-deep copy contructors
    public static void copyCommonFieldsByName(System.Object from, System.Object to)
    {
        var fieldsFrom = from.GetType().GetFields();

        for(int i=0; i<fieldsFrom.Count(); ++i){
            var fieldInfoTarget = to.GetType().GetField(fieldsFrom[i].Name, Util.ALL_METHOD_FLAGS);
            if (fieldInfoTarget != null)
                fieldInfoTarget.SetValue(to,fieldsFrom[i].GetValue(from));
        }
    }

    public static void replaceAllInstances(System.Type originalType, System.Type newType, string parentGameObjectName)
    {
        if ( !newType.IsSubclassOf(originalType) )
            return;

        var parent = GameObject.Find(parentGameObjectName);
        if(!parent)
            parent = new GameObject(parentGameObjectName);

        var toBeReplaced = Resources.FindObjectsOfTypeAll(originalType);
        for(int i=0; i<toBeReplaced.Count(); ++i){
            var replaceWith = parent.AddComponent(newType);
            copyCommonFieldsByName(toBeReplaced[i],replaceWith);
            
            //TODO: find all references to the replaced object and redirect them to the new (typ gc info)
        }

    }

}
